/*
 * DockIconController.m
 *
 * Copyright (c) 2006 Kent Sutherland
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to use,
 * copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the
 * Software, and to permit persons to whom the Software is furnished to do so,
 * subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
 * FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
 * COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
 * IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

#import "DockIconController.h"
#import "iChat.h"
#import "Chax.h"

static DockIconController *_sharedController = nil;

@interface NSBezierPath (RoundedRectAdditions)
+ (NSBezierPath*)bezierPathWithRoundRectInRect:(NSRect)aRect radius:(float)radius;
@end

@implementation NSBezierPath (RoundedRectAdditions)

+ (NSBezierPath*)bezierPathWithRoundRectInRect:(NSRect)aRect radius:(float)radius
{
   NSBezierPath* path = [self bezierPath];
   radius = MIN(radius, 0.5f * MIN(NSWidth(aRect), NSHeight(aRect)));
   NSRect rect = NSInsetRect(aRect, radius, radius);
   [path appendBezierPathWithArcWithCenter:NSMakePoint(NSMinX(rect), NSMinY(rect)) radius:radius startAngle:180.0 endAngle:270.0];
   [path appendBezierPathWithArcWithCenter:NSMakePoint(NSMaxX(rect), NSMinY(rect)) radius:radius startAngle:270.0 endAngle:360.0];
   [path appendBezierPathWithArcWithCenter:NSMakePoint(NSMaxX(rect), NSMaxY(rect)) radius:radius startAngle:  0.0 endAngle: 90.0];
   [path appendBezierPathWithArcWithCenter:NSMakePoint(NSMinX(rect), NSMaxY(rect)) radius:radius startAngle: 90.0 endAngle:180.0];
   [path closePath];
   return path;
}

@end

@implementation DockIconController

+ (DockIconController *)sharedController
{
	if (!_sharedController) {
		_sharedController = [[DockIconController alloc] init];
	}
	return _sharedController;
}

- (id)init
{
	if ( (self = [super init]) ) {
		_appIcon = [[NSImage imageNamed:@"NSApplicationIcon"] retain];
		[_appIcon setDelegate:self];
		_senders = [[NSMutableArray alloc] init];
		_counts = [[NSMutableDictionary alloc] init];
		_pictures = [[NSMutableDictionary alloc] init];
		_flashTimer = nil;
		_pictureIndex = 0;
		
		[[NSNotificationCenter defaultCenter] addObserver:self selector:@selector(notificationReceived:) name:NSApplicationDidBecomeActiveNotification object:nil];
		[[NSNotificationCenter defaultCenter] addObserver:self selector:@selector(notificationReceived:) name:NSApplicationWillTerminateNotification object:nil];
	}
	return self;
}

- (void)dealloc
{
	[[NSNotificationCenter defaultCenter] removeObserver:self];
	
	[_senders release];
	[_counts release];
	[_appIcon release];
	[super dealloc];
}

- (void)notificationReceived:(NSNotification *)notification
{
	if ([[notification name] isEqualToString:NSApplicationDidBecomeActiveNotification]) {
		[self clearUnreadMessages];
	} else if ([[notification name] isEqualToString:NSApplicationWillTerminateNotification]) {
		[NSApp setApplicationIconImage:_appIcon];
	}
}

- (void)newMessageFromSender:(NSString *)name picture:(NSImage *)picture
{
	if ([Chax boolForKey:@"EnableDockNotification"] || [Chax boolForKey:@"UseBuddyIconNotification"]) {
		int currentCount = 0;
		if (![_senders containsObject:name]) {
			[_senders addObject:name];
			if (picture) {
				[_pictures setObject:picture forKey:name];
			}
		} else {
			currentCount = [[_counts objectForKey:name] intValue];
		}
		
		if (!_flashTimer) {
			if ([Chax boolForKey:@"FlashBadge"]) {
				_flashTimer = [NSTimer scheduledTimerWithTimeInterval:1.0 target:self selector:@selector(flashFire) userInfo:nil repeats:YES];
			} else if ([Chax boolForKey:@"UseBuddyIconNotification"]) {
				_flashTimer = [NSTimer scheduledTimerWithTimeInterval:1.0 target:self selector:@selector(updateDockIcon) userInfo:nil repeats:YES];
			}
		}
		
		[_counts setObject:[NSNumber numberWithInt:currentCount + 1] forKey:name];
		[self updateDockIcon];
	}
}

- (void)removeSender:(NSString *)name
{
	if ([_senders containsObject:name]) {
		[_senders removeObject:name];
		[_counts removeObjectForKey:name];
		[_pictures removeObjectForKey:name];
		[self updateDockIcon];
	}
}

- (void)clearUnreadMessages
{
	[_senders removeAllObjects];
	[_counts removeAllObjects];
	[_pictures removeAllObjects];
	[self updateDockIcon];
	
	if (_flashTimer) {
		[_flashTimer invalidate];
	}
	_flashTimer = nil;
	_invert = NO;
}

- (void)flashFire
{
	_invert = !_invert;
	[self updateDockIcon];
}

- (void)updateDockIcon
{
	if (![Chax boolForKey:@"EnableDockNotification"] && ![Chax boolForKey:@"UseBuddyIconNotification"]) {
		return;
	}
	
	if ([_senders count] > 0) {
		NSImage *newIcon = [_appIcon copy];
		
		[newIcon lockFocus];
		
		if ([Chax boolForKey:@"UseBuddyIconNotification"]) {
			_pictureIndex++;
			_pictureIndex %= [_pictures count] + 1;
			if (_pictureIndex < [_pictures count]) {
				NSImage *picture = [_pictures objectForKey:[_senders objectAtIndex:_pictureIndex]];
				[picture drawInRect:NSMakeRect(0, 0, [newIcon size].width, [newIcon size].height) fromRect:NSMakeRect(0, 0, [picture size].width, [picture size].height) operation:NSCompositeSourceOver fraction:1.0];
			}
		}
		
		if ([Chax boolForKey:@"EnableDockNotification"]) {
			NSColor *foregroundColor, *backgroundColor;
			if ( !(foregroundColor = [NSUnarchiver unarchiveObjectWithData:[Chax dataForKey:@"DockForegroundColor"]]) ) {
				foregroundColor = [NSColor whiteColor];
			}
			
			if ( !(backgroundColor = [NSUnarchiver unarchiveObjectWithData:[Chax dataForKey:@"DockBackgroundColor"]]) ) {
				backgroundColor = [NSColor redColor];
			}
			
			if (_invert) {
				NSColor *tempColor = foregroundColor;
				foregroundColor = backgroundColor;
				backgroundColor = tempColor;
			}
			
			NSDictionary *attributes = [NSDictionary dictionaryWithObjectsAndKeys:[NSFont fontWithName:@"LucidaGrande-Bold" size:30], NSFontAttributeName, foregroundColor, NSForegroundColorAttributeName, nil, nil];
			
			if ([Chax integerForKey:@"DockNotificationType"] == 0) {
				NSString *string;
				NSSize size;
				int numMessages = 0;
				NSEnumerator *enumerator = [_counts objectEnumerator];
				NSNumber *nextNumber;
				
				while ( (nextNumber = [enumerator nextObject]) ) {
					numMessages += [nextNumber intValue];
				}
				string = [NSString stringWithFormat:@"%i", numMessages];
				
				size = [string sizeWithAttributes:attributes];
				
				[backgroundColor set];
				if ([Chax boolForKey:@"UseMailBadge"]) {
					NSImage *mailBadge = [[NSImage alloc] initByReferencingFile:[[NSBundle bundleWithPath:[[NSWorkspace sharedWorkspace] fullPathForApplication:@"Mail.app"]] pathForImageResource:@"newMailBadge1&2"]];
					NSSize size = [mailBadge size];
					[mailBadge drawInRect:NSMakeRect(80, 80, 48, 48) fromRect:NSMakeRect(0, 0, size.width, size.height) operation:NSCompositeSourceOver fraction:1.0];
					[mailBadge release];
				} else {
					[[NSBezierPath bezierPathWithOvalInRect:NSMakeRect(80, 80, 48, 48)] fill];
				}
				[string drawAtPoint:NSMakePoint(80 + ((48 - size.width) / 2), 86) withAttributes:attributes];
			} else {
				NSEnumerator *enumerator = [_senders objectEnumerator];
				NSString *nextString;
				float drawY = 96;
				
				while ( (nextString = [enumerator nextObject]) ) {
					[backgroundColor set];
					[[NSBezierPath bezierPathWithRoundRectInRect:NSMakeRect(0, drawY, 128, 32) radius:10.0] fill];
					[nextString drawInRect:NSMakeRect(5, drawY + 4, 123, 30) withAttributes:attributes];
					drawY -= 33;
				}
			}
		}
		
		[newIcon unlockFocus];
		[newIcon setDelegate:self];
		[NSApp setApplicationIconImage:[newIcon autorelease]];
	} else {
		[NSApp setApplicationIconImage:_appIcon];
	}
}

- (void)testDockIcon
{
	[NSTimer scheduledTimerWithTimeInterval:1.0 target:self selector:@selector(testTimeInterval:) userInfo:nil repeats:YES];
}

- (void)testTimeInterval:(NSTimer *)timer
{
	[self newMessageFromSender:[NSString stringWithFormat:@"Preview %i", [_senders count] + 1] picture:nil];
	if ([_senders count] > 5) {
		[timer invalidate];
		[self clearUnreadMessages];
	}
}

@end
