/*
 * HaxChatController.m
 *
 * Copyright (c) 2006 Kent Sutherland
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to use,
 * copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the
 * Software, and to permit persons to whom the Software is furnished to do so,
 * subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
 * FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
 * COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
 * IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

#import "HaxChatController.h"
#import "HaxChatInputLine.h"
#import "HaxAIMLGenerator.h"
#import "Chax.h"
#import "TabbedChatController.h"

static NSMutableSet *_tabbedChats = nil;
static NSMutableArray *_unloaded, *_unaccepted;

@implementation HaxChatController

+ (void)load
{
	_unloaded = [[NSMutableArray alloc] init];
	_unaccepted = [[NSMutableArray alloc] init];
}

- (void)acceptNotification:(id)fp8
{
	if ([Chax boolForKey:@"EnableTabs"]) {
		if (![self isLoaded] && ![[self chat] isKindOfClass:[SavedChat class]] && ![Chax boolForKey:@"SkipNewMessageNotification"] && ![[self chat] fromMe]) {
			[self setIsTab:NO];
			[_unloaded removeObject:self];
			return;
		}
		if ([_unaccepted containsObject:self]) {
			[_window setAlphaValue:0.0];
			[super acceptNotification:fp8];
			
			[_unloaded removeObject:self];
			[_unaccepted removeObject:self];
			
			[_window setAlphaValue:1.0];
			[_window orderOut:nil];
			if (![[TabbedChatController sharedController] containsChatController:self]) {
				[[TabbedChatController sharedController] addChatController:self chatView:_enclosingView drawer:_peopleDrawer];
			}
		}
		return;
	}
	
	if ([self isLoaded]) {
		[super acceptNotification:fp8];
	}
	
	if (([[self chat] isKindOfClass:[ActiveChat class]]) && [Chax boolForKey:@"SkipNewMessageNotification"]) {
		if ([Chax boolForKey:@"HideOnDeactivate"]) {
			[_window setHidesOnDeactivate:YES];
		}
		
		[super acceptNotification:fp8];
		
		[_unloaded removeObject:self];
		
		if (![self isTab]) {
			//Cascade the new window properly
			if (![Chax boolForKey:@"SkipCascading"]) {
				NSEnumerator *docs = [[NSApp orderedDocuments] objectEnumerator];
				NSRect frame = [_window frame];
				ChatDoc *nextDocument;
				
				while ( (nextDocument = [docs nextObject]) ) {
					HaxChatController *controller = [nextDocument controller];
					if ([controller isKindOfClass:[ChatController class]]) {
						NSRect nextFrame = [[controller window] frame];
						//Check for same origin as the new window
						if ((controller != self) && (frame.origin.x == nextFrame.origin.x) && (frame.origin.y == nextFrame.origin.y)) {
							frame.origin.x -= 23;
							frame.origin.y -= 23;
							[_window setFrame:frame display:NO];
						}
					}
				}
			}
			
			[_window makeKeyAndOrderFront:nil];
			
			[_window setLevel:NSStatusWindowLevel];
			[_window setLevel:NSNormalWindowLevel];
		}
		
		if ([Chax boolForKey:@"HideOnDeactivate"]) {
			[_window setHidesOnDeactivate:NO];
		}
	}
	
	[_unloaded removeObject:self];
	[_unaccepted removeObject:self];
	
	if ([[self chat] isComposing] || [[self chat] isKindOfClass:[SavedChat class]]) {
		[_window makeKeyAndOrderFront:nil];
	}
}

- (void)awakeFromNib
{
	if ([Chax boolForKey:@"EnableTabs"]) {
		[self setIsTab:YES];
	}
	
	[_unloaded addObject:self];
	[_unaccepted addObject:self];
	
	[super awakeFromNib];
	[(HaxChatInputLine *)_inputLine setSmileyButtonVisible:![Chax boolForKey:@"HideSmileyButton"]];
	
	if ([Chax boolForKey:@"EnableTabs"]) {
		[self performSelector:@selector(putIntoTab) withObject:nil afterDelay:0.0];
	}
	
	[self performSelector:@selector(acceptNotification:) withObject:nil afterDelay:0.0];
}

- (void)putIntoTab
{
	//Put a new window into a tab immediately if the chat is opened by the user, if the chat is a requested group chat, or if the chat is a saved chat
	if ([[self chat] fromMe] || [[self chat] isKindOfClass:[SavedChat class]] || [[self chat] joinState] == 2) {
		[[TabbedChatController sharedController] doNotSelectNextOff];
		[[TabbedChatController sharedController] addChatController:self chatView:_enclosingView drawer:_peopleDrawer];
	}
}

- (BOOL)isLoaded
{
	return ![_unloaded containsObject:self];
}

- (BOOL)isTab
{
	return [_tabbedChats containsObject:self];
}

- (void)setIsTab:(BOOL)flag
{
	if (_tabbedChats == nil) {
		_tabbedChats = [[NSMutableSet alloc] initWithCapacity:1];
	}
	
	if (flag) {
		[_tabbedChats addObject:self];
	} else {
		[_tabbedChats removeObject:self];
	}
}

- (NSDrawer *)peopleDrawer
{
	return _peopleDrawer;
}

- (NSView *)enclosingView
{
	return _enclosingView;
}

- (void)showParticipants:(id)fp8
{
	if ([self isTab]) {
		[_peopleDrawer setParentWindow:[[TabbedChatController sharedController] windowForChatController:self]];
	}
	[super showParticipants:fp8];
}

- (void)_showErrorSheet:(id)fp8
{
	if ([self isTab]) {
		NSBeginAlertSheet([NSString stringWithFormat:NSLocalizedString(@"%@ error:", nil), [[[self chat] service] shortName]], nil, nil, nil, [[TabbedChatController sharedController] windowForChatController:self], nil, nil, nil, nil, [fp8 objectForKey:@"ErrorMessage"]);
	} else {
		[super _showErrorSheet:fp8];
	}
}

- (void)addLink:(id)fp8
{
	if ([self isTab]) {
		[NSApp beginSheet:_linkSheet modalForWindow:[[TabbedChatController sharedController] windowForChatController:self] modalDelegate:self didEndSelector:@selector(_linkSheetDidEnd:returnCode:contextInfo:) contextInfo:NULL];
	} else {
		[super addLink:fp8];
	}
}

- (void)chooseChatOptions:(id)fp8
{
	if ([self isTab]) {
		[self _updateChatModeSheet];
		[NSApp beginSheet:_chatOptionsSheet modalForWindow:[[TabbedChatController sharedController] windowForChatController:self] modalDelegate:self didEndSelector:@selector(_chatOptionsDidEnd:returnCode:contextInfo:) contextInfo:NULL];
	} else {
		[super chooseChatOptions:fp8];
	}
}

- (void)_runInviteSheetForInvitees:(id)fp8
{
	if ([self isTab]) {
		[_inviteMsgLabel setStringValue:([fp8 count] == 1) ? [NSString stringWithFormat:NSLocalizedString(@"Enter invitation message for %@:", @"Enter invitation message for %@:"), [[fp8 objectAtIndex:0] name]] : NSLocalizedString(@"Enter invitation message:", @"Enter invitation message:")];
		
		[NSApp beginSheet:_inviteMsgSheet modalForWindow:[[TabbedChatController sharedController] windowForChatController:self] modalDelegate:self didEndSelector:@selector(_inviteMsgDidEnd:returnCode:contextInfo:) contextInfo:[fp8 retain]];
	} else {
		[super _runInviteSheetForInvitees:fp8];
	}
}

- (void)_chatOptionsDidEnd:(NSWindow *)sheet returnCode:(int)returnCode contextInfo:(void *)contextInfo
{
	[super _chatOptionsDidEnd:sheet returnCode:returnCode contextInfo:contextInfo];
	
	if ([self isTab]) {
		[[TabbedChatController sharedController] reloadLabels];
	}
}

- (void)_adjustInputLineHeightBy:(float)fp8
{
	if ([self isTab] && [self isInputLineVisible]) {
		//here we increase the size of the window by the requested about
		//change the autoresizing masks so that everything resizes properly,
		//then reset the masks so normal resizing works properly again
		NSWindow *tabWindow = [[TabbedChatController sharedController] windowForChatController:self];
		NSRect winFrame = [tabWindow frame], screenFrame = [[tabWindow screen] frame], visibleScreenFrame = [[tabWindow screen] visibleFrame];
		
		if (screenFrame.size.height < winFrame.size.height + fp8) {
			fp8 = screenFrame.size.height - winFrame.size.height;
		}
		
		[_inputLineTabView setAutoresizingMask:NSViewHeightSizable | NSViewWidthSizable | NSViewMinYMargin];
		
		winFrame.size.height += fp8;
		winFrame.origin.y -= fp8;
		
		if (winFrame.size.height > visibleScreenFrame.size.height) {
			winFrame.origin.y = visibleScreenFrame.origin.y;
			winFrame.size.height = visibleScreenFrame.size.height;
		} else if (winFrame.origin.y < screenFrame.origin.y) {
			winFrame.origin.y = screenFrame.origin.y;
		}
		
		[[[_transcript superview] superview] setAutoresizingMask:NSViewWidthSizable | NSViewMinYMargin];
		[tabWindow setFrame:winFrame display:YES animate:YES];
		[[[_transcript superview] superview] setAutoresizingMask:NSViewWidthSizable | NSViewHeightSizable];
		[_inputLineTabView setAutoresizingMask:NSViewWidthSizable | NSViewMaxYMargin];
	} else {
		[super _adjustInputLineHeightBy:fp8];
	}
}

- (void)_setInputLineHeight:(float)fp8
{
	if (![self isInputLineVisible]) {
		return;
	}
	
	//Do we have a minimum input line height set?
	float height = (float)[Chax integerForKey:@"MinimumInputLineHeight"];
	height = (height > fp8) ? height : fp8;
	
	if ([self isTab]) {
		if (height < 15) {
			height = 15;
		}
		[self _adjustInputLineHeightBy:height - [[self inputLine] frame].size.height + 7];
	} else {
		[super _setInputLineHeight:height];
	}
}

- (void)sendMessage:(id)sender
{
	//Mobile device warning
	if ([Chax boolForKey:@"WarnMobile"]) {
		Presentity *presentity = [[self chat] otherPerson];
		if ([presentity mobileDeviceName] != nil) {
			//Run a mobile user warning
			NSString *name = [presentity firstName];
			
			if (name == nil || [name length] == 0) {
				name = [presentity name];
			}
			
			NSBeginInformationalAlertSheet(LocalizedString(@"mobile_title"), nil, @"Cancel", nil, ([self isTab] ? [[TabbedChatController sharedController] windowForChatController:self] : [self window]), self, @selector(warningSheetDidEnd:returnCode:contextInfo:), NULL, sender, LocalizedString(@"mobile_msg"), [presentity name], name);
			return;
		}
	}
	
	/*if (1) {
		//Post the distributed notification for the Proxi trigger
		NSDictionary *info = [NSDictionary dictionaryWithObjectsAndKeys:
									@"MessageSent", @"Event",
									[[[self chat] otherPerson] name], @"User",
									@"", @"Message",
									[[[[self chat] otherPerson] image] TIFFRepresentation], @"Image",
									nil, nil];
		
		[[NSDistributedNotificationCenter defaultCenter] postNotificationName:@"ChaxProxiNotification" object:nil userInfo:info];
	}*/
	
	//If we're logged in through ICQ and we're talking to an ICQ user flag the AIML generator to send plain text
	if ([[[[self chat] otherPerson] ID] isICQorPhone] && ![Chax boolForKey:@"DisablePlainText"]) {
		[HaxAIMLGenerator nextMessagePlainText];
	}
	[super sendMessage:sender];
}

- (void)warningSheetDidEnd:(NSWindow *)sheet returnCode:(int)returnCode contextInfo:(void *)contextInfo
{
	if (returnCode == NSOKButton) {
		[super sendMessage:contextInfo];
	} else {
		[_inputLine selectText:nil];
	}
}

- (void)textView:(NSTextView *)aTextView doubleClickedOnCell:(id <NSTextAttachmentCell>)cell inRect:(NSRect)cellFrame atIndex:(unsigned)charIndex
{
	if ([Chax boolForKey:@"UsePreviewForImages"]) {
		NSString *imagePath = [[[cell attachment] fileWrapper] filename];
		NSDictionary *attributes = [[[cell attachment] fileWrapper] fileAttributes];
		
		if (imagePath) {
			NSDirectoryEnumerator *enumerator = [[NSFileManager defaultManager] enumeratorAtPath:NSTemporaryDirectory()];
			NSString *nextFile;
			BOOL found = NO;
			
			while ( !found && (nextFile = [enumerator nextObject]) ) {
				if ([[nextFile lastPathComponent] isEqualToString:imagePath]) {
					NSString *newPath = [NSTemporaryDirectory() stringByAppendingPathComponent:nextFile];
					NSDictionary *foundAttributes = [[NSFileManager defaultManager] fileAttributesAtPath:newPath traverseLink:NO];
					
					if ([[attributes objectForKey:@"NSFileSystemFileNumber"] isEqual:[foundAttributes objectForKey:@"NSFileSystemFileNumber"]]) {
						imagePath = newPath;
						found = YES;
					}
				}
			}
			
			if (!found) {
				imagePath = [NSTemporaryDirectory() stringByAppendingPathComponent:imagePath];
				if (![[NSFileManager defaultManager] fileExistsAtPath:imagePath]) {
					[[[cell attachment] fileWrapper] writeToFile:imagePath atomically:YES updateFilenames:NO];
				}
			}
			
			[[NSWorkspace sharedWorkspace] openFile:imagePath withApplication:@"Preview.app"];
		}
	}
}

@end
