/*
 * LogViewerController.m
 *
 * Copyright (c) 2006 Kent Sutherland
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to use,
 * copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the
 * Software, and to permit persons to whom the Software is furnished to do so,
 * subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
 * FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
 * COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
 * IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

#import "LogViewerController.h"
#import "Chax.h"
#import "UpdateController.h"
#import "AGRegex.h"
#import "LogFile.h"

static LogViewerController *_sharedController;

NSString *DeleteLogToolbarItemIdentifier = @"DeleteLogToolbarItemIdentifier";
NSString *ExportLogToolbarItemIdentifier = @"ExportLogToolbarItemIdentifier";
NSString *RevealLogToolbarItemIdentifier = @"RevealLogToolbarItemIdentifier";
NSString *SearchFieldToolbarItemIdentifier = @"SearchFieldToolbarItemIdentifier";

@implementation LogViewerController

+ (LogViewerController *)sharedController
{
	if (!_sharedController) {
		_sharedController = [[LogViewerController alloc] initWithWindowNibName:@"LogViewer"];
	}
	return _sharedController;
}

- (id)initWithWindowNibName:(NSString *)string
{
	if ( (self = [super initWithWindowNibName:string]) ) {
		_keys = [[NSMutableArray alloc] init];
		_logs = [[NSMutableDictionary alloc] init];
		_searchKeys = [[NSMutableArray alloc] init];
		_searchResults = [[NSMutableDictionary alloc] init];
		_firstLoadComplete = NO;
		
		[[NSNotificationCenter defaultCenter] addObserver:self selector:@selector(applicationLaunched:) name:NSApplicationDidFinishLaunchingNotification object:nil];
	}
	return self;
}

- (void)dealloc
{
	[[NSNotificationCenter defaultCenter] removeObserver:self];
	[_searchResults release];
	[_searchKeys release];
	[_toolbar release];
	[_searchField release];
	[_keys release];
	[_logs release];
	[super dealloc];
}

- (void)applicationLaunched:(NSNotification *)note
{
	if ([Chax boolForKey:@"AutoUpdateChax"]) {
		[[UpdateController sharedController] setAutomaticCheckingEnabled:YES];
		[[UpdateController sharedController] performSelector:@selector(checkForUpdates:) withObject:@"" afterDelay:2.0];
	}
	
	//[NSThread detachNewThreadSelector:@selector(reloadLogsThreaded) toTarget:self withObject:nil];
}

- (void)windowDidLoad
{
	[self setWindowFrameAutosaveName:@"LogViewer"];
	
	_toolbar = [[NSToolbar alloc] initWithIdentifier:@"LogToolbar"];
	[_toolbar setDelegate:self];
	[_toolbar setAllowsUserCustomization:YES];
	[_toolbar setAutosavesConfiguration:YES];
	[[self window] setToolbar:_toolbar];
	
	[_logTableView setTarget:self];
	[_logTableView setAction:@selector(tableViewAction:)];
	[_personTableView setTarget:self];
	[_logTableView reloadData];
	[_personTableView reloadData];
	
	[_chatTextView _commonInitialization];
	[_chatTextView setDelegate:self];
	
	Class rendererClass = NSClassFromString([Chax objectForKey:@"LogViewerRenderer"]);
	if (rendererClass) {
		MessageRenderer *renderer = [[[rendererClass alloc] initWithView:_chatTextView] autorelease];
		[_chatTextView setMessageRenderer:renderer];
	}
	
	int flags = [Chax integerForKey:@"LogViewerFlags"];
	if (flags > 0) {
		[_chatTextView setViewFlags:flags];
	}
	
	[[NSNotificationCenter defaultCenter] addObserver:self selector:@selector(tableViewNotificationReceived:) name:@"ChaxTableArrowDown" object:nil];
}

- (void)showWindow:(id)sender
{
	if (![[self window] isVisible]) {
		[NSThread detachNewThreadSelector:@selector(reloadLogsThreaded) toTarget:self withObject:NULL];
	}
	
	[super showWindow:sender];
	
	if (_badDate && !_ignoreDates) {
		NSBeginAlertSheet(LocalizedString(@"fix_log_dates"), LocalizedString(@"fix_dates"), LocalizedString(@"update_ignore"), nil, [self window], self, @selector(resetDatesSheetDidEnd:returnCode:contextInfo:), nil, nil, LocalizedString(@"fix_log_dates_msg"));
	}
}

- (void)resetDatesSheetDidEnd:(NSWindow *)sheet returnCode:(int)returnCode contextInfo:(void  *)contextInfo
{
	if (returnCode == NSOKButton) {
		[self performSelector:@selector(resetLogModificationDates) withObject:nil afterDelay:0.5];
	} else {
		_ignoreDates = YES;
	}
}

- (void)reloadLogsThreaded
{
	NSAutoreleasePool *pool = [[NSAutoreleasePool alloc] init];
	[self reloadLogs];
	[pool release];
}

- (void)reloadLogs
{
	[self createLogStructureFromEnumerator:[[NSFileManager defaultManager] enumeratorAtPath:[SavedChat savedChatDirectory]] inKeys:_keys inLogs:_logs];
	
	[_logTableView performSelectorOnMainThread:@selector(reloadData) withObject:nil waitUntilDone:NO];
	[_personTableView performSelectorOnMainThread:@selector(reloadData) withObject:nil waitUntilDone:NO];
	
	if (!_firstLoadComplete) {
		_firstLoadComplete = YES;
	}
}

- (void)createLogStructureFromEnumerator:(NSEnumerator *)enumerator inKeys:(NSMutableArray *)keys inLogs:(NSMutableDictionary *)logs
{
	//Regular expression that searches for a separator and then a date. This is the format of most logs
	AGRegex *primaryRegex = [AGRegex regexWithPattern:@" \\S{1,3} (([0-9]{4}|[0-9]{2}).){3}"];
	
	//Regular expression that searches for just a date
	AGRegex *secondaryRegex = [AGRegex regexWithPattern:@" (([0-9]{4}|[0-9]{2}).){3}"];
	
	//Regular expression that searches for the French date format
	AGRegex *tertiaryRegex = [AGRegex regexWithPattern:@" \\S{2,3} -(([0-9]{4}|[0-9]{2}).){3}"];
	
	NSString *nextFile;
	NSMutableArray *nextLogArray;
	
	[keys removeAllObjects];
	[logs removeAllObjects];
	
	_badDate = NO;
	
	while ( (nextFile = [enumerator nextObject]) ) {
		if ([[nextFile pathExtension] isEqualToString:@"chat"] || [[nextFile pathExtension] isEqualToString:@"ichat"]) {
			AGRegexMatch *match = [primaryRegex findInString:nextFile];
			NSString *name = nil;
			int location;
			
			//Search for matches on the log formats
			if (!match) {
				//Secondary search
				match = [secondaryRegex findInString:nextFile];
				
				if (!match) {
					//French format
					match = [tertiaryRegex findInString:nextFile];
				}
			}
			
			if (match) {
				location = [match range].location;
				name = [[nextFile substringToIndex:location] lastPathComponent];
				
				if (![keys containsObject:name]) {
					[keys addObject:[NSString stringWithUTF8String:[name cStringUsingEncoding:NSUTF8StringEncoding]]];
					[logs setObject:[NSMutableArray array] forKey:name];
				}
				
				//Add the log file to the list, naming it with the file creation date
				NSString *path = [[SavedChat savedChatDirectory] stringByAppendingPathComponent:nextFile];
				NSDictionary *attributes = [[NSFileManager defaultManager] fileAttributesAtPath:path traverseLink:NO];
				NSDate *creationDate = [attributes objectForKey:NSFileCreationDate];
				
				if ([creationDate earlierDate:[NSDate dateWithString:@"2000-01-01 12:00:00 +0000"]] == creationDate) {
					_badDate = YES;
				}
				
				NSString *fileName = [[SavedChat savedChatDirectory] stringByAppendingPathComponent:nextFile];
				NSString *logName = [creationDate descriptionWithCalendarFormat:@"%b %d, %Y, %1I:%M %p" timeZone:nil locale:[[NSUserDefaults standardUserDefaults] dictionaryRepresentation]];
				[[logs objectForKey:name] insertObject:[LogFile logFileWithName:logName fileName:fileName date:creationDate] atIndex:0];
			}
		}
	}
	
	//Sort all the inserte logs by date
	enumerator = [logs objectEnumerator];
	
	while ( (nextLogArray = [enumerator nextObject]) ) {
		[nextLogArray sortUsingSelector:@selector(compare:)];
	}
}

- (void)tableViewNotificationReceived:(NSNotification *)note
{
	if ([note object] == _logTableView) {
		[[self window] makeFirstResponder:_personTableView];
	} else {
		[[self window] makeFirstResponder:_logTableView];
		[_logTableView selectRow:0 byExtendingSelection:NO];
	}
}

- (void)showLogsForPresentity:(Presentity *)presentity
{
	if ([_keys count] == 0) {
		[self reloadLogs];
	}
	
	NSMutableString *name = [[presentity name] mutableCopy];
	CFStringNormalize((CFMutableStringRef)name, kCFStringNormalizationFormD);
	
	int index = [_keys indexOfObject:name]; //Search for the presentity requested
	if (index == NSNotFound) {
		index = [_keys indexOfObject:[presentity ID]]; //Search by screenname if nothing is returned
	}
	
	if (index == NSNotFound) {
		NSRunAlertPanel(LocalizedString(@"no_logs"), LocalizedString(@"no_logs_msg"), nil, nil, nil, [presentity name]);
	} else {
		[self showWindow:nil];
		[self performSelector:@selector(selectRow:) withObject:[NSNumber numberWithInt:index] afterDelay:0.0];
	}
	
	[name release];
}

- (void)selectRow:(NSNumber *)number
{
	int index = [number intValue];
	[_personTableView selectRow:index byExtendingSelection:NO];
	[_personTableView scrollRowToVisible:index];
}

- (void)tableViewAction:(id)sender
{
	NSArray *keys = [[_searchField stringValue] length] ? _searchKeys : _keys;
	NSDictionary *logs = [[_searchField stringValue] length] ? _searchResults : _logs;
	
	if ((sender == _logTableView) && ([keys count] > 0)) {
		if ([sender selectedRow] > -1) {
			NSString *name = [keys objectAtIndex:[_personTableView selectedRow]];
			NSString *file = [[[logs objectForKey:name] objectAtIndex:[sender selectedRow]] fileName];
			SavedChat *chat;
			
			if ([[file pathExtension] isEqualToString:@"ichat"]) {
				chat = [[SavedChat alloc] initWithTranscriptFile:file];
			} else {
				NSData *data = [[NSData alloc] initWithContentsOfFile:file];
				chat = [[SavedChat alloc] initWithSavedData:data];
				[data release];
			}
			
			[_chatTextView scrollRangeToVisible:NSMakeRange(0, 0)];
			[_chatTextView setChat:chat];
			[chat release];
		} else {
			[_chatTextView setChat:nil];
		}
	}
}

- (void)searchTextChanged:(id)sender
{
	//Clear out old search results
	[_searchResults removeAllObjects];
	[_searchKeys removeAllObjects];
	[_logTableView deselectAll:self];
	[_personTableView deselectAll:self];
	[_personTableView reloadData];
	[_logTableView reloadData];
	
	//Begin a search with the given search term, if any
	NSString *searchString = [sender stringValue];
	
	if ([searchString length] > 0) {
		if (_query) {
			[_query stopQuery];
			[_query release];
		}
		
		//Begin a new search with searchString
		_query = [[NSMetadataQuery alloc] init];
		[_query setSearchScopes:[NSArray arrayWithObject:[SavedChat savedChatDirectory]]];
		[_query setPredicate:[NSPredicate predicateWithFormat:@"kMDItemTextContent like[c] %@", searchString]];
		
		if ([_query startQuery]) {
			[[NSNotificationCenter defaultCenter] addObserver:self selector:@selector(spotlightNotificationReceived:) name:NSMetadataQueryDidFinishGatheringNotification object:_query];
		}
	} else {
		//Stop any current search
		if (_query) {
			[_query stopQuery];
			[_query release];
			_query = nil;
			[[NSNotificationCenter defaultCenter] removeObserver:self name:NSMetadataQueryDidFinishGatheringNotification object:_query];
		}
	}
}

- (void)spotlightNotificationReceived:(NSNotification *)note
{
	//Extract results out of the query
	int i;
	NSMutableArray *results = [[NSMutableArray alloc] init];
	
	for (i = 0; i < [[note object] resultCount]; i++) {
		NSString *filename = [[[note object] resultAtIndex:i] valueForAttribute:@"kMDItemFSName"];
		if (![results containsObject:filename]) {
			[results addObject:filename];
		}
	}
	
	//Build a structure out of the search results
	[self createLogStructureFromEnumerator:[results objectEnumerator] inKeys:_searchKeys inLogs:_searchResults];
	
	[_searchKeys sortUsingSelector:@selector(caseInsensitiveCompare:)];
	
	[results release];
	
	[_query stopQuery];
	[[NSNotificationCenter defaultCenter] removeObserver:self name:NSMetadataQueryDidFinishGatheringNotification object:_query];
	[_logTableView deselectAll:self];
	[_personTableView deselectAll:self];
	[_logTableView reloadData];
	[_personTableView reloadData];
}

- (void)setRendererFromMenuItem:(id)sender
{
	Class class = NSClassFromString([sender representedObject]);
	if (class) {
		MessageRenderer *renderer = [[[class alloc] initWithView:_chatTextView] autorelease];
		[_chatTextView setMessageRenderer:renderer];
		[Chax setObject:[sender representedObject] forKey:@"LogViewerRenderer"];
	}
}

- (void)setChatShowsNames:(id)sender
{
	int viewFlags = [_chatTextView viewFlags], flag = (1 << 1);
	
	if (viewFlags & 1) {
		viewFlags = ~viewFlags & 1;
	}
	
	if (!(viewFlags & flag)) {
		viewFlags |= flag;
	}
	
	[_chatTextView setViewFlags:viewFlags];
	[Chax setInteger:viewFlags forKey:@"LogViewerFlags"];
}

- (void)setChatShowsPictures:(id)sender
{
	int viewFlags = [_chatTextView viewFlags];
	
	if (viewFlags & (1 << 1)) {
		viewFlags = ~viewFlags & (1 << 1);
	}
	
	if (!(viewFlags & 1)) {
		viewFlags |= 1;
	}
	
	[_chatTextView setViewFlags:viewFlags];
	[Chax setInteger:viewFlags forKey:@"LogViewerFlags"];
}

- (void)setChatShowsNamesAndPictures:(id)sender
{
	int viewFlags = [_chatTextView viewFlags];
	
	viewFlags |= 1;
	viewFlags |= (1 << 1);
	
	[_chatTextView setViewFlags:viewFlags];
	[Chax setInteger:viewFlags forKey:@"LogViewerFlags"];
}

- (BOOL)validateMenuItem:(id <NSMenuItem>)menuItem
{
	NSString *representedObject = [menuItem representedObject];
	
	if (representedObject) {
		[menuItem setState:[representedObject isEqualToString:[[_chatTextView messageRenderer] className]] ? NSOnState : NSOffState];
	} else {
		int viewFlags = [_chatTextView viewFlags];
		
		if (([menuItem action] == @selector(setChatShowsNamesAndPictures:) && (viewFlags & (1 << 1)) && (viewFlags & 1)) || 
			([menuItem action] == @selector(setChatShowsNames:) && (viewFlags & (1 << 1)) && (~viewFlags & 1)) ||
			([menuItem action] == @selector(setChatShowsPictures:) && (viewFlags & 1) && (~viewFlags & (1 << 1)))) {
			[menuItem setState:NSOnState];
		} else {
			[menuItem setState:NSOffState];
		}
	}
	
	return YES;
}

#pragma mark -
#pragma mark Sheet End Methods
#pragma mark -

- (void)confirmDeleteSheetDidEnd:(NSWindow *)sheet returnCode:(int)returnCode contextInfo:(void  *)contextInfo
{
	if (returnCode == NSOKButton) {
		NSArray *keys = [[_searchField stringValue] length] ? _searchKeys : _keys;
		NSDictionary *logs = [[_searchField stringValue] length] ? _searchResults : _logs;
		NSString *name = [keys objectAtIndex:[_personTableView selectedRow]];
		
		if ([_logTableView numberOfSelectedRows] > 0) {
			//Remove a bunch of logs
			NSIndexSet *indexes = [_logTableView selectedRowIndexes];
			NSMutableArray *personLogs = [logs objectForKey:name];
			int i;
			
			for (i = 0; i < [indexes count]; i++) {
				//[[NSFileManager defaultManager] removeFileAtPath:[[personLogs objectAtIndex:[indexes firstIndex] + i] fileName] handler:nil];
				NSString *source = [[personLogs objectAtIndex:[indexes firstIndex] + i] fileName];
				[[NSWorkspace sharedWorkspace] performFileOperation:NSWorkspaceRecycleOperation source:[source stringByDeletingLastPathComponent] destination:@"" files:[NSArray arrayWithObject:[source lastPathComponent]] tag:nil];
			}
		} else {
			//Kill all logs from a single person
			NSEnumerator *logsToDelete = [[logs objectForKey:name] objectEnumerator];
			LogFile *nextLog;
			
			while ( (nextLog = [logsToDelete nextObject]) ) {
				//[[NSFileManager defaultManager] removeFileAtPath:[nextLog fileName] handler:nil];
				NSString *source = [nextLog fileName];
				[[NSWorkspace sharedWorkspace] performFileOperation:NSWorkspaceRecycleOperation source:[source stringByDeletingLastPathComponent] destination:@"" files:[NSArray arrayWithObject:[source lastPathComponent]] tag:nil];
			}
		}
		
		[_logTableView deselectAll:nil];
		[self reloadLogs];
		
		if ([_logTableView selectedRow] > -1) {
			NSString *name = [keys objectAtIndex:[_personTableView selectedRow]];
			[_chatTextView setChat:[[[SavedChat alloc] initWithTranscriptFile:[[[logs objectForKey:name] objectAtIndex:[_logTableView selectedRow]] fileName]] autorelease]];
		} else {
			[_chatTextView setChat:nil];
		}
	}
}

- (void)saveLogSheetDidEnd:(NSSavePanel *)sheet returnCode:(int)returnCode contextInfo:(void *)contextInfo
{
	if (returnCode == NSOKButton) {
		//Get a string representation of the currently selected log
		NSEnumerator *messages = [[_chatTextView messages] objectEnumerator];
		InstantMessage *nextMessage;
		NSMutableString *stringRepresentation = [[NSMutableString alloc] init];
		
		while ( (nextMessage = [messages nextObject]) ) {
			if ([nextMessage sender]) {
				[stringRepresentation appendFormat:@"%@ (%@): %@\n", [[nextMessage sender] name], [[nextMessage time] descriptionWithCalendarFormat:@"%H:%M:%S" timeZone:nil locale:nil], [[nextMessage text] string]];
			}/* else {
				[stringRepresentation appendFormat:@"%@\n", [[nextMessage text] string]];
			}*/
		}
		
		[stringRepresentation writeToFile:[sheet filename] atomically:YES];
		[stringRepresentation release];
	}
}

#pragma mark -
#pragma mark Modification Date Correction
#pragma mark -

- (void)resetLogModificationDates
{
	[_resetProgressIndicator setMaxValue:[[[NSFileManager defaultManager] directoryContentsAtPath:[SavedChat savedChatDirectory]] count]];
	[_resetProgressIndicator setDoubleValue:0.0];
	
	[NSApp beginSheet:_resetWindow modalForWindow:[self window] modalDelegate:self didEndSelector:nil contextInfo:nil];
	[NSThread detachNewThreadSelector:@selector(runLogDateReset) toTarget:self withObject:nil];
}

- (void)resetLogIncrementIndicator
{
	[_resetProgressIndicator incrementBy:1.0];
}

- (void)runLogDateReset
{
	NSAutoreleasePool *pool = [[NSAutoreleasePool alloc] init];
	NSEnumerator *enumerator = [[NSFileManager defaultManager] enumeratorAtPath:[SavedChat savedChatDirectory]];
	NSString *nextFile;
	
	while ( (nextFile = [enumerator nextObject]) ) {
		NSString *path = [[SavedChat savedChatDirectory] stringByAppendingPathComponent:nextFile];
		SavedChat *chat = nil;
		
		if ([[path pathExtension] isEqualToString:@"ichat"]) {
			chat = [[SavedChat alloc] initWithTranscriptFile:path];
		} else if ([[path pathExtension] isEqualToString:@"chat"]) {
			NSData *data = [[NSData alloc] initWithContentsOfFile:path];
			chat = [[SavedChat alloc] initWithSavedData:data];
			[data release];
		}
		
		if (chat) {
			[_resetTextField performSelectorOnMainThread:@selector(setStringValue:) withObject:nextFile waitUntilDone:NO];
			[[NSFileManager defaultManager] changeFileAttributes:[NSDictionary dictionaryWithObject:[chat dateCreated] forKey:NSFileCreationDate] atPath:path];
		}
		
		[self performSelectorOnMainThread:@selector(resetLogIncrementIndicator) withObject:nil waitUntilDone:NO];
	}
	
	[NSApp endSheet:_resetWindow];
	[_resetWindow orderOut:nil];
	[self performSelectorOnMainThread:@selector(reloadLogs) withObject:nil waitUntilDone:NO];
	
	[pool release];
}

#pragma mark -
#pragma mark ChatTextView Delegate
#pragma mark -

- (void)textView:(NSTextView *)aTextView doubleClickedOnCell:(id <NSTextAttachmentCell>)cell inRect:(NSRect)cellFrame atIndex:(unsigned)charIndex
{
	if ([Chax boolForKey:@"UsePreviewForImages"]) {
		NSString *imagePath = [[[cell attachment] fileWrapper] preferredFilename];
		if (imagePath) {
			NSDirectoryEnumerator *enumerator = [[NSFileManager defaultManager] enumeratorAtPath:NSTemporaryDirectory()];
			NSString *nextFile;
			BOOL found = NO;
			
			while ( !found && (nextFile = [enumerator nextObject]) ) {
				if ([[nextFile lastPathComponent] isEqualToString:imagePath]) {
					imagePath = [NSTemporaryDirectory() stringByAppendingPathComponent:nextFile];
					found = YES;
				}
			}
			
			if (!found) {
				imagePath = [NSTemporaryDirectory() stringByAppendingPathComponent:imagePath];
				if (![[NSFileManager defaultManager] fileExistsAtPath:imagePath]) {
					[[[cell attachment] fileWrapper] writeToFile:imagePath atomically:YES updateFilenames:NO];
				}
			}
			
			[[NSWorkspace sharedWorkspace] openFile:imagePath withApplication:@"Preview.app"];
		}
	}
}

#pragma mark -
#pragma mark Toolbar Delegate
#pragma mark -

- (NSArray *)toolbarAllowedItemIdentifiers:(NSToolbar *)toolbar {
	return [NSArray arrayWithObjects:DeleteLogToolbarItemIdentifier, ExportLogToolbarItemIdentifier, RevealLogToolbarItemIdentifier, SearchFieldToolbarItemIdentifier, NSToolbarFlexibleSpaceItemIdentifier, NSToolbarSpaceItemIdentifier, NSToolbarSeparatorItemIdentifier, nil];
}

- (NSArray *)toolbarDefaultItemIdentifiers:(NSToolbar *)toolbar {
	return [NSArray arrayWithObjects:DeleteLogToolbarItemIdentifier, NSToolbarSeparatorItemIdentifier, ExportLogToolbarItemIdentifier, RevealLogToolbarItemIdentifier, NSToolbarFlexibleSpaceItemIdentifier, SearchFieldToolbarItemIdentifier, nil];
}

- (NSToolbarItem *)toolbar:(NSToolbar *)toolbar itemForItemIdentifier:(NSString *)itemIdentifier willBeInsertedIntoToolbar:(BOOL)flag {
	NSToolbarItem *item = [[NSToolbarItem alloc] initWithItemIdentifier:itemIdentifier];
	
	[item setAction:@selector(toolbarItemAction:)];
	[item setTarget:self];
	
	if ([itemIdentifier isEqualToString:DeleteLogToolbarItemIdentifier]) {
		NSImage *deleteImage = [[NSImage alloc] initByReferencingFile:[[NSBundle bundleWithPath:[[NSWorkspace sharedWorkspace] fullPathForApplication:@"Mail.app"]] pathForImageResource:@"delete.tiff"]];
		[item setLabel:LocalizedString(@"delete")];
		[item setPaletteLabel:LocalizedString(@"delete")];
		[item setImage:deleteImage];
		[deleteImage release];
	} else if ([itemIdentifier isEqualToString:ExportLogToolbarItemIdentifier]) {
		NSImage *exportImage = [[NSImage alloc] initByReferencingFile:[[NSBundle bundleWithPath:[[NSWorkspace sharedWorkspace] fullPathForApplication:@"TextEdit.app"]] pathForImageResource:@"txt.icns"]];
		[item setLabel:LocalizedString(@"export_as_text")];
		[item setPaletteLabel:LocalizedString(@"export_as_text")];
		[item setImage:exportImage];
		[exportImage release];
	} else if ([itemIdentifier isEqualToString:RevealLogToolbarItemIdentifier]) {
		NSImage *finderIcon = [[NSImage alloc] initByReferencingFile:[[NSBundle bundleWithPath:[[NSWorkspace sharedWorkspace] fullPathForApplication:@"Finder.app"]] pathForImageResource:@"Finder.icns"]];
		[item setLabel:LocalizedString(@"reveal_in_finder")];
		[item setPaletteLabel:LocalizedString(@"reveal_in_finder")];
		[item setImage:finderIcon];
		[finderIcon release];
	} else if ([itemIdentifier isEqualToString:SearchFieldToolbarItemIdentifier]) {
		[item setLabel:LocalizedString(@"search")];
		[item setPaletteLabel:LocalizedString(@"search")];
		
		if (!_searchField) {
			_searchField = [[NSSearchField alloc] initWithFrame:NSMakeRect(0, 0, 22, 50)];
			[_searchField setTarget:self];
			[_searchField setAction:@selector(searchTextChanged:)];
		}
		
		[item setView:_searchField];
		
		[item setMaxSize:NSMakeSize(200, 24)];
		[item setMinSize:NSMakeSize(50, 24)];
		
		[item setImage:nil];
	} else {
		[item release];
		item = nil;
	}
	
	return [item autorelease];
}

- (void)toolbarItemAction:(id)sender {
	if ([[sender itemIdentifier] isEqualToString:DeleteLogToolbarItemIdentifier]) {
		NSString *name = [_keys objectAtIndex:[_personTableView selectedRow]];
		if ([[self window] firstResponder] == _personTableView && [_personTableView selectedRow] > -1 || [_logTableView numberOfSelectedRows] > 1) {
			//Prompt to delete a whole bunch of logs
			NSBeginAlertSheet(LocalizedString(@"delete_logs"), LocalizedString(@"delete"), LocalizedString(@"dont_delete"), nil, [self window], self, NULL, @selector(confirmDeleteSheetDidEnd:returnCode:contextInfo:) , nil, [NSString stringWithFormat:LocalizedString(([[self window] firstResponder] == _personTableView && [_personTableView selectedRow] > -1) ? @"delete_all_logs_msg" : @"delete_selected_logs_msg"), name]);
		} else if ([_logTableView selectedRow] > -1) {
			NSArray *keys = [[_searchField stringValue] length] ? _searchKeys : _keys;
			NSDictionary *logs = [[_searchField stringValue] length] ? _searchResults : _logs;
			
			//[[NSFileManager defaultManager] removeFileAtPath:[[[logs objectForKey:name] objectAtIndex:[_logTableView selectedRow]] fileName] handler:nil];
			NSString *source = [[[logs objectForKey:name] objectAtIndex:[_logTableView selectedRow]] fileName];
			[[NSWorkspace sharedWorkspace] performFileOperation:NSWorkspaceRecycleOperation source:[source stringByDeletingLastPathComponent] destination:@"" files:[NSArray arrayWithObject:[source lastPathComponent]] tag:nil];
			
			if ([_logTableView selectedRow] == [[logs objectForKey:name] count] - 1) {
				//The last row is selected, so select one row above
				[_logTableView selectRow:[_logTableView selectedRow] - 1 byExtendingSelection:NO];
			}
			
			[self reloadLogs];
			
			if ([_logTableView selectedRow] > -1) {
				NSString *name = [keys objectAtIndex:[_personTableView selectedRow]];
				[_chatTextView setChat:[[[SavedChat alloc] initWithTranscriptFile:[[[logs objectForKey:name] objectAtIndex:[_logTableView selectedRow]] fileName]] autorelease]];
			} else {
				[_chatTextView setChat:nil];
			}
		}
	} else if ([[sender itemIdentifier] isEqualToString:ExportLogToolbarItemIdentifier]) {
		NSSavePanel *panel = [NSSavePanel savePanel];
		[panel setCanSelectHiddenExtension:YES];
		[panel setCanCreateDirectories:YES];
		[panel setRequiredFileType:@"txt"];
		[panel beginSheetForDirectory:nil file:nil modalForWindow:[self window] modalDelegate:self didEndSelector:@selector(saveLogSheetDidEnd:returnCode:contextInfo:) contextInfo:nil];
	} else if ([[sender itemIdentifier] isEqualToString:RevealLogToolbarItemIdentifier]) {
		if ([_logTableView selectedRow] > -1) {
			NSArray *keys = [[_searchField stringValue] length] ? _searchKeys : _keys;
			NSDictionary *logs = [[_searchField stringValue] length] ? _searchResults : _logs;
			
			NSString *name = [keys objectAtIndex:[_personTableView selectedRow]];
			NSString *path = [[[logs objectForKey:name] objectAtIndex:[_logTableView selectedRow]] fileName];
			[[NSWorkspace sharedWorkspace] selectFile:path inFileViewerRootedAtPath:[path stringByDeletingLastPathComponent]];
		}
	} else if ([[sender itemIdentifier] isEqualToString:SearchFieldToolbarItemIdentifier]) {
		//Do nothing.
	}
}

- (void)toolbarDidRemoveItem:(NSNotification *)notification
{
	NSEnumerator *itemEnumerator = [[[notification object] items] objectEnumerator];
	NSToolbarItem *nextItem;
	
	while ( (nextItem = [itemEnumerator nextObject]) ) {
		if ([[nextItem itemIdentifier] isEqualToString:SearchFieldToolbarItemIdentifier]) {
			return;
		}
	}
	
	[_searchField setStringValue:@""];
	[self searchTextChanged:_searchField];
}

- (BOOL)validateToolbarItem:(NSToolbarItem *)item
{
	if ([[item itemIdentifier] isEqualToString:DeleteLogToolbarItemIdentifier]) {
		return [_personTableView selectedRow] > -1;
	} else if ([[item itemIdentifier] isEqualToString:ExportLogToolbarItemIdentifier] || [[item itemIdentifier] isEqualToString:RevealLogToolbarItemIdentifier]) {
		return [_logTableView selectedRow] > -1;
	}
	return YES;
}

#pragma mark -
#pragma mark TableView Data Source
#pragma mark -

- (int)numberOfRowsInTableView:(NSTableView *)tableView
{
	NSArray *keys = [[_searchField stringValue] length] ? _searchKeys : _keys;
	NSDictionary *logs = [[_searchField stringValue] length] ? _searchResults : _logs;
	
	int number = [keys count];
	if (tableView == _logTableView) {
		int row = [_personTableView selectedRow];
		number = (number == 0) ? 1 : ((row > -1) ? [[logs objectForKey:[keys objectAtIndex:row]] count] : 0);
	}
	return number;
}

- (id)tableView:(NSTableView *)tableView objectValueForTableColumn:(NSTableColumn *)aTableColumn row:(int)rowIndex
{
	NSArray *keys = ([[_searchField stringValue] length] && ![_toolbar customizationPaletteIsRunning]) ? _searchKeys : _keys;
	NSDictionary *logs = ([[_searchField stringValue] length] && ![_toolbar customizationPaletteIsRunning]) ? _searchResults : _logs;
	
	if (tableView == _personTableView) {
		return [keys objectAtIndex:rowIndex];
	} else {
		NSString *value = nil;
		if ([keys count] == 0) {
			if ([_keys count] == 0) {
				value = LocalizedString(@"loading_logs");
			} else {
				value = (_query && ![_query isStopped]) ? LocalizedString(@"searching") : LocalizedString(@"no_logs");
			}
		} else {
			int row = [_personTableView selectedRow];
			if (row > -1) {
				NSArray *logList = [logs objectForKey:[keys objectAtIndex:row]];
				if ([logList count] > rowIndex) {
					value = (row > -1) ? [[logList objectAtIndex:rowIndex] name] : nil;
				}
			}
		}
		return value;
	}
}

- (BOOL)tableView:(NSTableView *)tableView shouldSelectRow:(int)rowIndex
{
	return _firstLoadComplete;
}

- (void)tableViewSelectionDidChange:(NSNotification *)note
{
	id sender = [note object];
	NSArray *keys = [[_searchField stringValue] length] ? _searchKeys : _keys;
	NSDictionary *logs = [[_searchField stringValue] length] ? _searchResults : _logs;
	
	if (sender == _personTableView) {
		[_logTableView deselectAll:self];
		[_logTableView reloadData];
		[_chatTextView setChat:nil];
	} else if ((sender == _logTableView) && ([keys count] > 0)) {
		if ([sender selectedRow] > -1) {
			NSString *name = [keys objectAtIndex:[_personTableView selectedRow]];
			NSString *file = [[[logs objectForKey:name] objectAtIndex:[sender selectedRow]] fileName];
			SavedChat *chat;
			
			if ([[file pathExtension] isEqualToString:@"ichat"]) {
				chat = [[SavedChat alloc] initWithTranscriptFile:file];
			} else {
				NSData *data = [[NSData alloc] initWithContentsOfFile:file];
				chat = [[SavedChat alloc] initWithSavedData:data];
				[data release];
			}
			
			[_chatTextView setChat:chat];
			[chat release];
		} else {
			[_chatTextView setChat:nil];
		}
	}
}

@end
