/*
 * UpdateController.m
 *
 * Copyright (c) 2006 Kent Sutherland
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to use,
 * copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the
 * Software, and to permit persons to whom the Software is furnished to do so,
 * subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
 * FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
 * COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
 * IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

#import "UpdateController.h"
#import "Chax.h"
#import <SystemConfiguration/SystemConfiguration.h>

#define UPDATE_SITE [NSURL URLWithString:@"http://www.ksuther.com/chax/"]
#define UPDATE_URL [NSURL URLWithString:@"http://www.ksuther.com/chax/version.php?n"]
#define UPDATE_INTERVAL 864000

static UpdateController *_sharedController = nil;

@implementation UpdateController

+ (UpdateController *)sharedController
{
	if (_sharedController == nil) {
		_sharedController = [[UpdateController alloc] init];
	}
	return _sharedController;
}

- (id)init
{
	if ( (self = [super init]) ) {
		_updateTimer = nil;
		_showNoUpdates = YES;
		_updateResponseData = [[NSMutableData alloc] init];
	}
	return self;
}

- (void)dealloc
{
	[[NSNotificationCenter defaultCenter] removeObserver:self];
	[_updateResponseData release];
	[super dealloc];
}

- (void)setAutomaticCheckingEnabled:(BOOL)flag
{
	if (flag) {
		_updateTimer = [NSTimer scheduledTimerWithTimeInterval:UPDATE_INTERVAL target:self selector:@selector(checkForUpdates:) userInfo:nil repeats:YES];
	} else if (_updateTimer != nil) {
		[_updateTimer invalidate];
		_updateTimer = nil;
	}
}

- (void)checkForUpdates:(id)sender
{
	NSTimeInterval timeSinceLastUpdate = [Chax integerForKey:@"LastUpdate"];
	
	if (sender == nil || timeSinceLastUpdate == 0 || [[NSDate dateWithTimeIntervalSinceReferenceDate:timeSinceLastUpdate] timeIntervalSinceNow] < -UPDATE_INTERVAL + 3600) {
		//Check for network reachability
		const char *host = "ksuther.com";
		BOOL reachable, result = 0;
		SCNetworkConnectionFlags flags = 0;
		reachable = SCNetworkCheckReachabilityByName(host, &flags);
		_showNoUpdates = (sender == nil);
		if (reachable && ((flags & kSCNetworkFlagsReachable) && !(flags & kSCNetworkFlagsConnectionRequired)))  {
			[_updateResponseData setData:[NSData data]];
			[[NSURLConnection connectionWithRequest:[NSURLRequest requestWithURL:UPDATE_URL cachePolicy:NSURLRequestReloadIgnoringCacheData timeoutInterval:15] delegate:self] retain];
		} else if (_showNoUpdates) {
			NSRunAlertPanel(LocalizedString(@"updateerror_title"), LocalizedString(@"updateerror_msg"), nil, nil, nil, LocalizedString(@"updateerror_msg_unreachable"));
		}
	}
}

#pragma mark -
#pragma mark NSURLConnection handler methods
#pragma mark -

- (void)connection:(NSURLConnection *)connection didFailWithError:(NSError *)error
{
	if (_showNoUpdates) {
		NSRunAlertPanel(LocalizedString(@"updateerror_title"), LocalizedString(@"updateerror_msg"), nil, nil, nil, [error localizedDescription]);
	}
	[_updateResponseData setData:[NSData data]];
	[connection release];
}

- (void)connection:(NSURLConnection *)connection didReceiveData:(NSData *)data
{
	[_updateResponseData appendData:data];
}

- (void)connectionDidFinishLoading:(NSURLConnection *)connection
{
	NSString *string = [[NSString alloc] initWithData:_updateResponseData encoding:NSASCIIStringEncoding];
	
	NSArray *lines = [string componentsSeparatedByString:@"\n"];
	NSString *currentVersionString = [[[NSBundle bundleForClass:[self class]] infoDictionary] objectForKey:@"CFBundleVersion"];
	[_updateResponseData setData:[NSData data]];
	
	if ([[lines objectAtIndex:0] doubleValue] > [[[[NSBundle bundleWithIdentifier:@"com.ksuther.chaxloader"] infoDictionary] objectForKey:@"VersionNumber"] doubleValue]) {
		if (NSRunInformationalAlertPanel(LocalizedString(@"update_title"), LocalizedString(@"update_msg"), LocalizedString(@"update_get"), LocalizedString(@"update_ignore"), nil, currentVersionString, [lines objectAtIndex:1]) == NSOKButton) {
			[[NSWorkspace sharedWorkspace] openURL:UPDATE_SITE];
		}
	} else if (_showNoUpdates) {
		NSRunInformationalAlertPanel(LocalizedString(@"noupdate_title"), LocalizedString(@"noupdate_msg"), nil, nil, nil, currentVersionString);
	}
	[connection release];
	
	[Chax setInteger:[NSDate timeIntervalSinceReferenceDate] forKey:@"LastUpdate"];
}

@end